"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = exports.ERROR_FILE_EXIST = exports.ERROR_NO_SUCH_FILE = exports.PKG_FILE_NAME = exports.TABLE_NAME = void 0;

var _path = _interopRequireDefault(require("path"));

var _stream = require("stream");

var _lib = require("@verdaccio/commons-api/lib");

var _streams = require("@verdaccio/streams");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

const TABLE_NAME = 'files';
exports.TABLE_NAME = TABLE_NAME;
const PKG_FILE_NAME = 'package.json';
exports.PKG_FILE_NAME = PKG_FILE_NAME;
const ERROR_NO_SUCH_FILE = new Error('no such file');
exports.ERROR_NO_SUCH_FILE = ERROR_NO_SUCH_FILE;
ERROR_NO_SUCH_FILE.code = 'ENOENT';
const ERROR_FILE_EXIST = new Error('file exist');
exports.ERROR_FILE_EXIST = ERROR_FILE_EXIST;
ERROR_FILE_EXIST.code = 'EEXISTS';

const noop = () => {// Make linter happy
};

class PGPackageManager {
  constructor(sql, prefix, logger) {
    _defineProperty(this, "prefix", void 0);

    _defineProperty(this, "logger", void 0);

    _defineProperty(this, "sql", void 0);

    _defineProperty(this, "table", void 0);

    this.sql = sql;
    this.prefix = prefix;
    this.logger = logger;
    this.table = this._createTable(TABLE_NAME);
  }

  async updatePackage(name, updateHandler, onWrite, transformPackage, onEnd) {
    this.logger.debug({
      name
    }, '[pg-storage/updatePackage]: update package init ${name}');

    try {
      const buffer = await this._readStorageFile(this._getStorage(PKG_FILE_NAME));
      const json = JSON.parse(buffer.toString('utf8'));
      updateHandler(json, err => {
        if (err) {
          this.logger.error({
            err
          }, '[pg-storage/updatePackage/updateHandler]: onEnd @{err}');
          onEnd(err);
        } else {
          const transformed = transformPackage(json);
          this.logger.debug({
            transformed
          }, '[pg-storage/updatePackage/updateHandler]: onWrite @{transformed}');
          onWrite(name, transformed, onEnd);
        }
      });
    } catch (err) {
      this.logger.error({
        err
      }, '[pg-storage/updatePackage/updateHandler]: onEnd catch @{err}');
      return onEnd(err);
    }

    return;
  }

  async deletePackage(packageName, callback) {
    this.logger.debug({
      packageName
    }, '[local-storage/deletePackage] delete a package @{packageName}');

    try {
      await this._deleteFile(this._getStorage(packageName));
      callback(null);
    } catch (err) {
      callback(err);
    }
  }

  async removePackage(callback) {
    this.logger.debug({
      packageName: this.prefix
    }, '[pg-storage/removePackage] remove a package: @{packageName}');

    try {
      await this._deletePrefix(this._getStorage('.'));
      callback(null);
    } catch (err) {
      callback(err);
    }
  }

  createPackage(name, value, callback) {
    this.logger.debug({
      packageName: name
    }, '[pg-storage/createPackage] create a package: @{packageName}');

    this._createFile(this._getStorage(PKG_FILE_NAME), this._convertToString(value), callback);
  }

  savePackage(name, value, callback) {
    this.logger.debug({
      packageName: name
    }, '[pg-storage/savePackage] save a package: @{packageName}');

    this._writeFile(this._getStorage(PKG_FILE_NAME), Buffer.from(this._convertToString(value)), callback);
  }

  async readPackage(name, callback) {
    this.logger.debug({
      packageName: name
    }, '[pg-storage/readPackage] read a package: @{packageName}');

    try {
      const file = await this._readStorageFile(this._getStorage(PKG_FILE_NAME));
      const data = JSON.parse(file.toString('utf8'));
      this.logger.trace({
        packageName: name
      }, '[pg-storage/readPackage/_readStorageFile] read a package succeed: @{packageName}');
      callback(null, data);
    } catch (err) {
      this.logger.trace({
        err
      }, '[local-storage/readPackage/_readStorageFile] error on read a package: @{err}');
      callback(err);
    }
  }

  writeTarball(name) {
    const uploadStream = new _streams.UploadTarball({});
    this.logger.debug({
      packageName: name
    }, '[pg-storage/writeTarball] write a tarball for package: @{packageName}');
    let ended = false;
    uploadStream.on('end', function () {
      ended = true;
    });

    const pathName = this._getStorage(name);

    (async () => {
      try {
        await this._readStorageFile(pathName);
        uploadStream.emit('error', (0, _lib.getCode)(409, ERROR_FILE_EXIST.message));
      } catch (err) {
        if (err != ERROR_NO_SUCH_FILE) {
          uploadStream.emit('error', err);
        }
      }

      const chunks = [];
      uploadStream.on('data', data => chunks.push(data));

      uploadStream.done = () => {
        const query = async () => {
          const buffer = Buffer.concat(chunks);
          const err = await this._writeFile(pathName, buffer);

          if (err) {
            uploadStream.emit('error', err);
          } else {
            uploadStream.emit('success');
          }
        };

        if (ended) {
          query();
        } else {
          uploadStream.on('end', query);
        }
      };

      uploadStream.abort = noop;
      uploadStream.emit('open');
    })();

    return uploadStream;
  }

  readTarball(name) {
    const pathName = this._getStorage(name);

    this.logger.debug({
      packageName: name
    }, '[pg-storage/readTarball] read a tarball for package: @{packageName}');
    const readTarballStream = new _streams.ReadTarball({});

    const query = async () => {
      try {
        const file = await this._readStorageFile(pathName);
        readTarballStream.emit('content-length', file.byteLength);
        readTarballStream.emit('open');

        _stream.Readable.from(file).pipe(readTarballStream);
      } catch (err) {
        readTarballStream.emit('error', err);
      }
    };

    query();
    return readTarballStream;
  }

  async _createFile(name, contents, callback) {
    await this.table;
    this.logger.trace({
      name
    }, '[pg-storage/_createFile] create a new file: @{name}');

    try {
      await this._readStorageFile(name);
      this.logger.trace({
        name
      }, '[pg-storage/_createFile] file cannot be created, it already exists: @{name}');
      return callback((0, _lib.getCode)(409, 'EEXIST'));
    } catch (err) {
      if (err == ERROR_NO_SUCH_FILE) {
        this._writeFile(name, Buffer.from(contents), callback);
      }
    }
  }

  async _readStorageFile(name) {
    await this.table;
    this.logger.trace({
      name
    }, '[pg-storage/_readStorageFile] read a file: @{name}');

    try {
      const rows = await this.sql`SELECT content FROM ${this.sql(TABLE_NAME)} WHERE path = ${name}`;

      if (rows.count === 0) {
        throw ERROR_NO_SUCH_FILE;
      }

      const [{
        content: data
      }] = rows;
      this.logger.trace({
        name
      }, '[pg-storage/_readStorageFile] read file succeed: @{name}');
      return data;
    } catch (err) {
      this.logger.trace({
        name
      }, '[pg-storage/_readStorageFile] error on read the file: @{name}');
      throw err;
    }
  }

  _convertToString(value) {
    return JSON.stringify(value, null, '\t');
  }

  _getStorage(fileName = '') {
    const storagePath = _path.default.join(this.prefix, fileName);

    return storagePath;
  }

  async _writeFile(dest, data, cb = noop) {
    await this.table;

    try {
      await this.sql.begin(async sql => {
        await sql`
          INSERT INTO ${sql(TABLE_NAME)} (path, content, updated_at) VALUES (${dest}, ${data}, NOW())
          ON CONFLICT (path) DO UPDATE SET content = ${Buffer.from(data)}
        `;
      });
      cb(null);
      return null;
    } catch (err) {
      cb(err);
      return err;
    }
  }

  async _deleteFile(name) {
    await this.table;
    await this.sql`
      DELETE FROM ${this.sql(TABLE_NAME)} WHERE path = ${name}
    `;
  }

  async _deletePrefix(prefix) {
    await this.table;
    await this.sql`
      DELETE FROM ${this.sql(TABLE_NAME)} WHERE path = ${prefix + '%'}
    `;
  }

  async _createTable(name) {
    await this.sql`
        CREATE TABLE IF NOT EXISTS ${this.sql(name)} (
        path TEXT PRIMARY KEY,
        content bytea,
        created_at timestamp not null default current_timestamp,
        updated_at timestamp not null default current_timestamp
      )`;
    return name;
  }

}

exports.default = PGPackageManager;